from util import *
try:
    import ldap
    import ldapurl
    have_ldap = True
except:
    have_ldap = False

import subprocess
from subprocess import PIPE
from psycopg2 import IntegrityError
import syslog

# Send sensitive authentication logs to syslog
# but use this for other stuff.  We want to keep
# all user and password information out of the 
# publically readable log file.
log = logging.getLogger("cumin.authenticator")


class _syslog(object):
    log_auth = False
    @classmethod
    def log(cls, msg):
        if _syslog.log_auth:
            syslog.syslog(msg)

class CuminAuthenticatorLDAP(object):
    def __init__(self, url):
        log.info("Initializing %s", self)
        myurl = url.split("=",1)[1]
        try:
            self.url = ldapurl.LDAPUrl(myurl)
            log.info("Authenticator: parsed LDAP url successfully.")
        except Exception,e:
            log.error("Authenticator: cannot parse LDAP url %s", e)
    
    def __prep_con(self):
        try:
            conn = ldap.initialize(self.url.urlscheme + "://" +self.url.hostport)
        except Exception,e:
            log.error("Authenticator: cannot contact ldap server %s",e)
            return False

        if self.url.who and self.url.cred:
            try:
                conn.simple_bind_s(self.url.who,self.url.cred)
            except:
                log.error("Authenticator: cannot authenticate as service %s", 
                          self.url.who)
                return False
        return conn

    def change_pw(self,user,oldpass,newpass):
        conn = self.__prep_con()
        if not conn:
            return False
        filter = self.url.filterstr % user
        try:
            res = conn.search_s(self.url.dn, self.url.scope, filter)
        except Exception, e:
            log.error("Authenticator: update password, "\
                      "query returned exception %s", e)
            res = []
        if not res:
           log.info("Authenticator: update password, "\
                    "query returned no results in %s", self.__class__.__name__)
        else:
            dn,ent = res[0]
            conn.simple_bind_s(dn,oldpass)
            conn.passwd_s(dn,oldpass,newpass)
            log.info("Authenticator: update password succeeded in %s", 
                     self.__class__.__name__)
            _syslog.log("cumin: updated password via LDAP for user %s" % dn)
            return True
        return False

    def batch_import(self):
        log.debug("Authenticator: batch import in %s", self.__class__.__name__)
        userlist = res = []
        conn = self.__prep_con()
        error = not conn
        if not error:
            filter = self.url.filterstr % '*'
            try:
                res = conn.search_s(self.url.dn, self.url.scope, filter)
            except Exception, e:
                log.error("Authenticator:, batch import, "\
                          "query returned exception %s", e) 
                error = True
        if not res:
            log.info("Authenticator: batch import failed, "\
                     "query returned no results in %s", self.__class__.__name__)
        else:
            for dn,ent in res:
                dn_arr = ldap.dn.explode_rdn(dn)
                naming_attr = dn_arr[0].split('=')[0]
                userlist.append(ent[naming_attr][0])
        return userlist, error

    def authenticate(self, username, password):
        log.debug("Authenticating against %s", self.__class__.__name__)
        conn = self.__prep_con()
        if not conn:
            return False
        filter = self.url.filterstr % username
        try:
            res = conn.search_s(self.url.dn, self.url.scope, filter)
        except Exception, e:
            log.error("Authenticator: authenticate, "\
                      "query returned exception %s", e)
            res = []
        if not res:
            log.info("Authenticator: authentication failed, "\
                     "query returned no results in %s", self.__class__.__name__)
        else:
            for dn,ent in res:
                try:
                    conn.simple_bind_s(dn,password)
                    _syslog.log("cumin: authenticated user %s via LDAP" % dn)
                    return True
                except Exception, e:
                    log.info("Authenticator: LDAP authentication "\
                             "returned exception %s" % e)
                    _syslog.log("cumin: authentication via LDAP failed "\
                                "for user" % dn)
        return False

class CuminAuthenticatorScript(object):
    def __init__(self, commandline):
        self.params = {}
        self.cap = []
        me = self.__class__.__name__
        for a in commandline.split(','):
            param,val = a.split('=')
            self.params[param] = val

        if self.params.has_key('script') and \
           os.access(self.params['script'], os.X_OK):
            if self.params.has_key('auth'):
                log.debug("Authenticator: adding auth capability to %s", me)
                self.cap.append('a')
            if self.params.has_key('change'):
                log.debug("Authenticator: adding change capability to %s", me)
                self.cap.append('c')
            if self.params.has_key('list'):
                log.debug("Authenticator: adding list capability to %s", me)
                self.cap.append('l')
        else:
            log.error('Authenticator: script parameter missing or target file '\
                      'not executable in %s', me)

    def authenticate(self, username, password):
        me = self.__class__.__name__
        if 'a' in self.cap:
            log.debug("Authenticator: executing %s in %s", 
                      self.params['script'], me)
            # Use nameless pipes to pass parameters.  
            # If they are passed as arguments they show up in the output 
            # of ps, etc.  This way is secure.
            # Note, the call to communicate causes an EOF on stdin
            cmd = [self.params['script']]
            res = subprocess.Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            out, err = res.communicate(input="%s %s %s" % \
                                       (self.params['auth'], username, password))
            log.debug("Authenticator: authenticate, script returned "\
                          "%s, %s, exitcode %s", out, err, res.returncode)
            if res.returncode == 0:
                _syslog.log("cumin: authenticated user %s via script" % username)
                return True
        _syslog.log("cumin: authentication via script failed for user %s" % username)
        return False

    def batch_import(self):
        if 'l' in self.cap:
            log.debug("Authenticator: batch import in %s",
                      self.__class__.__name__)
            # Use nameless pipes to pass parameters.  
            # If they are passed as arguments they show up in the output 
            # of ps, etc.  This way is secure.
            # Note, the call to communicate causes an EOF on stdin
            cmd = [self.params['script']]
            res = subprocess.Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            out, err = res.communicate(input=self.params['list'])
            log.debug("Authenticator: batch import, script returned "\
                      "%s, %s, exitcode %s", out, err, res.returncode)
            return out.split(','), False
        return [], False

    def change_pw(self, user, newpass, oldpass):
        if 'c' in self.cap:
            me = self.__class__.__name__
            log.debug("Authenticator: change password in %s", me)

            # Use nameless pipes to pass parameters.  
            # If they are passed as arguments they show up in the output 
            # of ps, etc.  This way is secure.
            # Note, the call to communicate causes an EOF on stdin
            cmd = [self.params['script']]
            try:
                res = subprocess.Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
                out, err = res.communicate(input="%s %s %s %s" % \
                                           (self.params['change'], user, newpass, oldpass))
                log.debug("Authenticator: update password, script returned "\
                          "%s, %s, exitcode %s", out, err, res.returncode)
            except Exception, e:
                import traceback
                traceback.print_exc()
                res.returncode = -1

            if res.returncode == 0:
                _syslog.log("cumin: updated password via script for user %s" % user)
                return True
        return False

if have_ldap:
    authen_map = {'ldap': CuminAuthenticatorLDAP,
                  'script': CuminAuthenticatorScript}
else:
    authen_map = {'script': CuminAuthenticatorScript}

class CuminAuthenticator(object):
    def __init__(self, app, log_authentication=True):
        self.authenticators=[]
        self.app = app
        _syslog.log_auth = log_authentication
        log.info("Initializing %s", self)
        if not have_ldap:
            log.info("Authenticator: import of ldap modules failed, "\
                     "ldap mechanism not available")
        for authmech in app.authmech:
            if authmech != "internal":
                try:
                    mechname = authmech.split('=')[0]
                    log.info("Authenticator: adding auth mechanism %s", mechname)
                    classname = authen_map[mechname]
                except KeyError:
                    log.error("Authenticator: unsupported authenticator type %s",
                              mechname)
                    continue
                self.authenticators.append(classname(authmech))

    def batch_import(self, backend, output_on=False):
        numusers = 0
        numconflicts = 0
        errors = False
        import_list = []
        log.debug("Authenticator: calling batch import")
        try:
            classname = authen_map[backend]
        except KeyError:
            log.error("Authenticator: batch import, unsupported authenticator "\
                      "type '%s'", backend)
            errors = True

        if not errors:
            try:
                idx = [x.__class__.__name__ for x in self.authenticators]
                idx = idx.index(classname.__name__)
            except ValueError:
                log.error("Authenticator: batch import, authenticator "\
                          "not configured '%s'", backend)
                errors = True

        if not errors:
            users, errors = self.authenticators[idx].batch_import()
            for imp in users:
                import_list.append(imp)

        if len(import_list) > 0:
            try:
                conn = self.app.database.get_connection()
                cursor = conn.cursor()
                role = self.app.admin.get_role(cursor, "user")
                for importuser in import_list:
                    try:
                        user = self.app.admin.add_user(cursor, importuser, "")
                        if output_on:
                            print("Importing user %s" % importuser)
                        self.app.admin.add_assignment(cursor, user, role)
                        conn.commit()
                        numusers += 1
                    except IntegrityError:
                        if output_on:
                            print("Import failed, a user called '%s' "\
                                      "already exists" % importuser)
                        numconflicts += 1
                        conn.rollback()                        
            finally:
                conn.close()
        return numusers, numconflicts, errors

    def authenticate(self, username, password):
        log.debug("Authenticator: calling authenticate")
        cursor = self.app.database.get_read_cursor()
        cls = self.app.model.com_redhat_cumin.User
        user = cls.get_object(cursor, name=username)
        external = len(user.password) == 0
        log.info("Authenticator: authenticating external user %s" % external)
        if not external:
            res = crypt(password, user.password) == user.password
            if res:
                msg = "cumin: authenticated user %s" % username
            else:
                msg = "cumin: authentication failed for user %s" % username
            _syslog.log(msg)
            return res
        else:
            for authenticator in self.authenticators:
                log.debug("Authenticator: authenticate, try %s", authenticator)
                if authenticator.authenticate(username, password):
                    return True
        return False

    def update_password(self, username, oldpassword, newpassword):
        status = False
        message = ""
        log.debug("Authenticator: calling update_password")
        conn = self.app.database.get_connection()
        try:
            cursor = conn.cursor()
            cls = self.app.model.com_redhat_cumin.User
            user = cls.get_object(cursor, name=username)
            if user.password == "":
                for authenticator in self.authenticators:
                    status = authenticator.authenticate(username, oldpassword)
                    if status:
                        status = authenticator.change_pw(username, 
                                                        oldpassword, newpassword)
                        if not status:
                            message = "Could not change password"
                        break

                if not status and message == "":
                    message = "The password is incorrect"
            else:
                status = crypt(oldpassword, user.password) == user.password
                if status:
                    user.password = crypt_password(newpassword)
                    user.save(cursor)
                    conn.commit()
                    _syslog.log("cumin: updated password for user %s" % username)
                else:
                    message = "The password is incorrect"
        finally:
            conn.close()
        return status, message
