import os
import threading
import logging
import random
import urllib2
import socket
import string
import time
import sage

from datetime import datetime
from suds import *
from suds.client import Client
from suds.transport.https import HttpAuthenticated
from sage.util import CallSync, CallThread, ObjectPool, host_list
from sage.https import *
try:
    from sage.https_full import HTTPSFullCertTransport
    has_full_cert = True
    technology = sage.https_full.technology
except:
    has_full_cert = False

log = logging.getLogger("sage.aviary")

class SudsLogging(object):
    _on = False
    sudslogs = {"suds.client": None, "suds.transport": None, "suds.xsd.schema": None, "suds.wsdl": None}

    @classmethod
    def set(cls, flag, home):
        if flag:
            SudsLogging._on = True
            for k,v in SudsLogging.sudslogs.iteritems():
                l = logging.getLogger(k)
                l.setLevel(logging.DEBUG)
                try:
                    f = open(os.path.join(home, "log/"+k+".log"), 'a+')
                    h = logging.StreamHandler(f)
                    l.addHandler(h)
                    SudsLogging.sudslogs[k] = h
                except:
                    pass
            log.debug("AviaryOperations: suds logging on")

        elif SudsLogging._on:
            SudsLogging._on = False
            for k,v in SudsLogging.sudslogs.iteritems():
                try:
                    logging.getLogger(k).removeHandler(v)
                except:
                    pass
            log.debug("AviaryOperations: suds logging off")


#Stuff to do:

#- Test pool/schedd/stuff with hierarchical collectors

#- Add a summary comment at the top like QMF opreations

#- can we use default/timeout with suds?
#looks like the way to do this is through the Transport object that
#is set in the client.  The timeout is set on the Transport itself,
#and the Transport can be changed on the client with set_options.
#Alternatively, we could retain a reference to the Transport in 
#the clients that we pool so that we can call set_options on the
#transport.

class _AviaryJobMethods(object):

    # Do this here rather than __init__ so we don't have to worry about
    # matching parameter lists in multiple inheritance cases with super
    def init(self, job_servers, datadir):

        # Replace any occurrence of locahost with output of gethostname()
        # before parsing to match Machine fields of QMF objects later on.
        host = socket.gethostname()
        job_servers = string.replace(job_servers, "localhost", host)

        self.job_servers = host_list(job_servers, 
                                     default_scheme = "http", 
                                     default_port="9090",
                                     default_path="/services/job/")
                                                   
        job_wsdl = "file:" + os.path.join(datadir, "aviary-job.wsdl")
        self.job_client_pool = JobClientPool(job_wsdl, None)

    def set_job_attribute(self, scheduler, job_id, name, value, callback, submission):
        assert callback
        
        def my_callback(result):
            self.job_client_pool.return_object(job_client)
            result = self._pretty_result(result, scheduler.Machine)
            # massage results for use by standard callback
            cb_args = self._cb_args_dataless(result)
            callback(*cb_args)
                     
        job_client = self._setup_client(self.job_client_pool,
                                        scheduler.Machine,
                                        self.job_servers,
                                        "setJobAttribute")
                                   
        # Make a job id parameter (see job wsdl)
        jobId = job_client.factory.create('ns0:JobID')
        jobId.job = job_id
        jobId.pool = scheduler.Pool
        jobId.scheduler = scheduler.Name
        jobId.submission.name = submission.Name
        jobId.submission.owner = submission.Owner

        # Make attribute parameter from name and value
        aviary_attr = job_client.factory.create('ns0:Attribute')
        aviary_attr.name = name
        aviary_attr.type = "STRING"
        aviary_attr.value = value

        t = CallThread(job_client.service.setJobAttribute, my_callback, jobId, aviary_attr)
        t.start()

    def submit_job(self, scheduler, ad, callback):
        assert callback

        def my_callback(result):
            # Turn this back off before we put it back in the pool
            # so allow_overrides isn't set for someone else...
            job_client.set_enable_attributes(False)
            self.job_client_pool.return_object(job_client)
            result = self._pretty_result(result, scheduler.Machine)
            if isinstance(result, Exception):
                callback(result, None)                
            else:
                # the aviary response has the job id available,
                # we'll pass it anyway even though Cumin does not care
                # at the present time
                status = _AviaryCommon._get_status(result.status)
                if status == "OK" and hasattr(result, "id"):
                    id = result.id
                else:
                    id = None
                callback(status, id)

        job_client = self._setup_client(self.job_client_pool,
                                        scheduler.Machine,
                                        self.job_servers,
                                        "submitJob")

        # Set basic attributes in the order defined by aviary-job.wsdl.
        args = list()
        basic_attrs = ("Cmd", "Args", "Owner", "Iwd", "Submission")
        for attr in basic_attrs:
            args.append(ad[attr])

        # Add empty list for Aviary's basic requirement value...
        args.append([])

        # and let's let Requirements remain an unrestricted expression so that
        # we can just pass through the value from Cumin without interfering.
        # To do that, we need to specify Requirements through the
        # "extras" fields and set allowOverrides to True.
        # (otherwise, Requirements will be limited to particular
        # resource constraint types defined by aviary)
        job_client.set_enable_attributes(True)
        job_client.set_attributes({"allowOverrides": True})
        extras = list()
        for k, v in ad.iteritems():
            # We don't need to send descriptors down to aviary
            # and basic_attrs have already been filled in
            if k == "!!descriptors" or k in basic_attrs:
                continue
            
            extra = job_client.factory.create('ns0:Attribute')
            extra.name = k
            # But we do need to look in descriptors to find expressions...
            if k in ad["!!descriptors"]:
                extra.type = "EXPRESSION"
            else:
                try:
                    extra.type = self.type_to_aviary[type(v)]
                except KeyError:
                    extra.type = "UNDEFINED"
            extra.value = v
            extras.append(extra)

        # Important, extras itself must be added as an embedded list or 
        # suds will consider only a single item
        args.append(extras)

        t = CallThread(job_client.service.submitJob, my_callback, *args)
        t.start()

    def hold_job(self, scheduler, job_id, reason, submission, *args, **kwargs):
        '''
        This method is asynchronous iff 'callback' is supplied.

        kwargs will be searched for 'callback', 'default' and 'timeout' arguments.

        These optional arguments were moved to kwargs for compatibility
        with another implementation of the same routine.
        '''

        callback = "callback" in kwargs and kwargs["callback"] or None
        default = "default" in kwargs and kwargs["default"] or None
        timeout = "timeout" in kwargs and kwargs["timeout"] or 5
        
        return self._control_job(scheduler, job_id, reason, submission,
                                 "holdJob",
                                 callback, default, timeout)

    def release_job(self, scheduler, job_id, reason, submission, *args, **kwargs):
        '''
        This method is asynchronous iff 'callback' is supplied.

        kwargs will be searched for 'callback', 'default' and 'timeout' arguments.

        These optional arguments were moved to kwargs for compatibility
        with another implementation of the same routine.
        '''

        callback = "callback" in kwargs and kwargs["callback"] or None
        default = "default" in kwargs and kwargs["default"] or None
        timeout = "timeout" in kwargs and kwargs["timeout"] or 5
        
        return self._control_job(scheduler, job_id, reason, submission,
                                 "releaseJob",
                                 callback, default, timeout)

    def remove_job(self, scheduler, job_id, reason, submission, *args, **kwargs):
        '''
        This method is asynchronous iff 'callback' is supplied.

        kwargs will be searched for 'callback', 'default' and 'timeout' arguments.

        These optional arguments were moved to kwargs for compatibility
        with another implementation of the same routine.
        '''

        callback = "callback" in kwargs and kwargs["callback"] or None
        default = "default" in kwargs and kwargs["default"] or None
        timeout = "timeout" in kwargs and kwargs["timeout"] or 5

        return self._control_job(scheduler, job_id, reason, submission,
                                 "removeJob",
                                 callback, default, timeout)

    def _control_job(self, scheduler, job_id, reason, submission,
                     meth_name,
                     callback, default, timeout):

        client = self._setup_client(self.job_client_pool,
                                    scheduler.Machine,
                                    self.job_servers,
                                    meth_name)

        meth = getattr(client.service, meth_name)

        # Make a job id parameter (see job wsdl)
        jobId = client.factory.create('ns0:JobID')
        jobId.job = job_id
        jobId.pool = scheduler.Pool
        jobId.scheduler = scheduler.Name
        jobId.submission.name = submission.Name
        jobId.submission.owner = submission.Owner

        if callback:
            def my_callback(result):
                self.job_client_pool.return_object(client)
                # Fix up the exception message if necessary
                result = self._pretty_result(result, scheduler.Machine)
                cb_args = self._cb_args_dataless(result)
                callback(*cb_args)

            t = CallThread(meth, my_callback, jobId, reason)
            t.start()
        else:
            def my_process_results(result):
                # Fix up the exception message if necessary
                result = self._pretty_result(result, scheduler.Machine)
                return self._cb_args_dataless(result)

            res = self._call_sync(my_process_results, meth, jobId, reason)            
            self.job_client_pool.return_object(client)
            return res;

class _AviaryQueryMethods(object):

    # Do this here rather than __init__ so we don't have to worry about
    # matching parameter lists in multiple inheritance cases with super
    def init(self, query_servers, datadir):

        # Replace any occurrence of locahost with output of gethostname()
        # before parsing to match Machine fields of QMF objects later on.
        host = socket.gethostname()
        query_servers = string.replace(query_servers, "localhost", host)

        self.query_servers = host_list(query_servers, 
                                       default_scheme = "http", 
                                       default_port="9091",
                                       default_path="/services/query/")

        query_wsdl = "file:" + os.path.join(datadir, "aviary-query.wsdl")
        self.query_client_pool = QueryClientPool(query_wsdl, None)

    def fetch_job_data(self, job_server, job_id, ftype, file, start, end,
                       scheduler_name, submission, *args, **kwargs):
        '''
        kwargs will be searched for 'default' and 'timeout' arguments.

        These optional arguments were moved to kwargs for compatibility
        with another implementation of the same routine.
        '''

        default = "default" in kwargs and kwargs["default"] or None
        timeout = "timeout" in kwargs and kwargs["timeout"] or 5

        # Aviary doesn't use the file name as does QMF, instead it
        # specifies the file type and lets condor figure out the path.

        def my_process_results(result):
            # Fix up the exception message if necessary
            result = self._pretty_result(result, job_server.Machine)
            if isinstance(result, Exception):
                status = result
                data = None
            else:
                status = _AviaryCommon._get_status(result.status)
                if status == "OK" and hasattr(result, "content"):
                    # Match the format expected by Cumin.  This is
                    # the format used by the QMF call...
                    data = {'Data': result.content}
                else:
                    data = None
            return (status, data)

        client = self._setup_client(self.query_client_pool,
                                    job_server.Machine,
                                    self.query_servers,
                                    "getJobData")

        # Make a job data parameter (see query wsdl)
        jobData = client.factory.create('ns0:JobData')
        jobData.id.job = job_id
        jobData.id.pool = job_server.Pool
        jobData.id.scheduler = scheduler_name
        jobData.id.submission.name = submission.Name
        jobData.id.submission.owner = submission.Owner

        # Translate cumin file type to Aviary file type
        if ftype == "e":
            jobData.type = "ERR"
        elif ftype == "o":
            jobData.type = "OUT"
        elif ftype == "u":
            jobData.type = "LOG"
        else:
            # We can't translate the type.
            # Let Aviary throw an error on this instead of us.
            jobData.type = ftype

        from_end = start < 0
        max_bytes = abs(end - start)

        res = self._call_sync(my_process_results, client.service.getJobData, 
                              jobData, max_bytes, from_end )            
        self.query_client_pool.return_object(client)
        return res;

    def get_job_ad(self, job_server, job_id, scheduler_name, submission, 
                   *args, **kwargs):
        '''
        kwargs will be searched for 'default' and 'timeout' arguments.

        These optional arguments were moved to kwargs for compatibility
        with another implementation of the same routine.
        '''        

        default = "default" in kwargs and kwargs["default"] or None
        timeout = "timeout" in kwargs and kwargs["timeout"] or 5

        def make_tuple(attr):
            # Attempt to cast the value into the specified type
            if attr.type in self.aviary_to_type:
                try:
                    v = self.aviary_to_type[attr.type](attr.value)
                except:
                    v = attr.value
            else:
                v = attr.value
            return (attr.name, v)
        
        def make_dict(attrs):
            return dict([make_tuple(attr) for attr in attrs])

        def my_process_results(result):
            # Fix up the exception message if necessary
            result = self._pretty_result(result, job_server.Machine)
            if isinstance(result, Exception):
                status = result
                data = None
            else:
                status = _AviaryCommon._get_status(result[0].status)
                if status == "OK":
                    # Match the format expected by Cumin.  This is
                    # the format used by the QMF call.  We have a list
                    # of attributes in attrs that we need to make into
                    # a dictionary
                    ads = make_dict(result[0].details.attrs) 
                    data = {'JobAd': ads}
                else:
                    data = None
            return (status, data)

        client = self._setup_client(self.query_client_pool,
                                    job_server.Machine,
                                    self.query_servers,
                                    "getJobDetails")

        # Make a job id parameter (see job wsdl)
        jobId = client.factory.create('ns0:JobID')
        jobId.job = job_id
        jobId.pool = job_server.Pool
        jobId.scheduler = scheduler_name
        jobId.submission.name = submission.Name
        jobId.submission.owner = submission.Owner

        res = self._call_sync(my_process_results, 
                              client.service.getJobDetails, jobId)            
        self.query_client_pool.return_object(client)
        return res;

    def get_job_summaries(self, submission, callback, machine_name):
        assert callback

        def to_int_seconds(dt):
            # Change a datetime.datetime into int seconds since epoch
            # Note, this works nicely if the datetime happens to include microseconds
            # since the call to timetuple will drop them.  Stuff coming back from
            # condor should not have microseconds anyway.
            return int(time.mktime(dt.timetuple()))

        def get_string(job, attr):
            # Cast suds text types into str so we have standard Py types
            # Handles optional strings as well
            if hasattr(job, attr):
                return str(getattr(job, attr))
            return ""

        def adapt(jobs):
            # Make an aviary job summary look like the canonical form
            # that cumin is expecting (actually the QMF form because of history).
            result = list()
            for job in jobs:
                cluster, proc = job.id.job.split(".")
                j = dict()
                j["ClusterId"]            = int(cluster)
                j["Cmd"]                  = str(job.cmd)
                j["EnteredCurrentStatus"] = to_int_seconds(job.last_update)

                # Note, GlobalJobId here will not match the same value from
                # QMF because the qdate portion of the name is missing
                j["GlobalJobId"]          = job.id.scheduler + \
                                            "#" + job.id.job

                j["JobStatus"]            = str(job.job_status)
                j["ProcId"]               = int(proc)
                j["QDate"]                = to_int_seconds(job.queued)
                
                # These may be null...
                j["Args"]                 = get_string(job, "args1")
                j["ReleaseReason"]        = get_string(job, "released")
                j["HoldReason"]           = get_string(job, "held")
                result.append(j)
            return result
                
        def my_callback(result):
            query_client.set_enable_attributes(False)
            self.query_client_pool.return_object(query_client)
            result = self._pretty_result(result, machine_name)
            if isinstance(result, Exception):
                callback(result, None)                
            else:
                status =  _AviaryCommon._get_status(result[0].status)
                if status == "OK" and hasattr(result[0], "jobs"):
                    data = {"Jobs": adapt(result[0].jobs)}
                else:
                    data = {"Jobs": None}
                callback(status, data)

        query_client = self._setup_client(self.query_client_pool,
                                          machine_name,
                                          self.query_servers,
                                          "getSubmissionSummary")

        # What we really want here is the job summaries from the
        # submission summary response.  To get those, we have to
        # set an extra attribute on the client...
        query_client.set_enable_attributes(True)
        query_client.set_attributes({"includeJobSummaries": "true"})

        # Make a submission id.  (see query wsdl)
        subId = query_client.factory.create('ns0:SubmissionID')
        subId.name = submission.Name
        subId.owner = submission.Owner

        t = CallThread(query_client.service.getSubmissionSummary, 
                       my_callback, subId)
        t.start()


class _AviaryCommon(object):
    def __init__(self, name, key="", cert="", root_cert="", domain_verify=True):
        self.name = name
        self.type_to_aviary = self._type_to_aviary()
        self.aviary_to_type = self._aviary_to_type()

        self.key = key
        self.cert = cert
        self.root_cert = root_cert
        self.domain_verify = domain_verify
        self.server_validation_possible = has_full_cert
        if self.root_cert == "":
            log.info("AviaryOperations: no root certificate file specified, "\
                     "using client validation only for ssl connections.")

        elif not self.server_validation_possible:
            log.info("AviaryOperations: server certificate validation not "\
                     "supported, using client validation "\
                     "only for ssl connections.")
        else:
            log.info("AviaryOperations: using client and server "\
                     "certificate validation for ssl connections, "\
                     "solution is %s" % technology)

            log.info("AviaryOperations: verify server domain against "\
                     "certificate during validation (%s)" % self.domain_verify)

    @classmethod
    def _type_to_aviary(cls):
        # Need to be able to turn simple Python types into Aviary types for attributes
        return {int: "INTEGER", float: "FLOAT", str: "STRING", bool: "BOOLEAN"}

    @classmethod
    def _aviary_to_type(cls):
        return {"INTEGER": int, "FLOAT": float, "STRING": str, "BOOLEAN": bool}

    @classmethod
    def _get_host(cls, name, servers):
        scheme = ""
        host = ""
        if name in servers:
            urls = servers[name]
            if len(urls) > 0:
                url = random.sample(urls, 1)[0]
                scheme = url.scheme
                host = str(url)
                # A particular method name is going to be appended to path,
                # so ensure the file "/" here.  Since we supply default path
                # values when the host list is parsed, we know the last portion
                # of the host string has to be a path.
                if not host.endswith("/"):
                    host += "/"
        return scheme, host

    @classmethod
    def _raise_no_host(cls, name):
        raise Exception("No aviary job servers specified for %s, check config files" % name)

    def _setup_client(self, client_pool, host_name, host_list, method):
        # Get a client object from the client_pool.  Use host_name as
        # a lookup in the host_list to find an initial URL and append the
        # method name.  Set up the client to point at that URL and return.
        client = client_pool.get_object()
        scheme, host = self._get_host(host_name, host_list)
        if host == "":
            self._raise_no_host(host_name)
        # Have to set the URL for the method.  This might go away someday...
        client.set_options(location=host+method)
        # Since we pool the clients and reuse them for different requests
        # and since its possible to be using servers with different schemes,
        # we have to always reset the transport here.
        if scheme == "https":
            if not os.path.isfile(self.key):
                raise Exception("Private key file "\
                                "for ssl communication with Aviary not found")
            if not os.path.isfile(self.cert):
                raise Exception("Client certificate file "\
                                "for ssl communication with Aviary not found")
            if self.root_cert != "" and self.server_validation_possible:
                if not os.path.isfile(self.root_cert):
                    raise Exception("Root certificate file "\
                                    "for Aviary server validation not found")
                the_transport = HTTPSFullCertTransport(self.key, 
                                                       self.cert, 
                                                       self.root_cert,
                                                       self.domain_verify)
            else:
                the_transport = HTTPSClientCertTransport(self.key, self.cert)
        else:
            # this is the default transport when none is specified
            the_transport = HttpAuthenticated()
        client.set_options(transport=the_transport)
        return client

    @classmethod
    def _get_status(cls, result):
        # For Aviary operations, if the operation
        # did not work the reason is in the text field.
        # In cumin, we want to pass any error text as
        # the status parameter to callbacks
        if result.code != "OK":
            return result.text
        return result.code

    @classmethod
    def _cb_args_dataless(cls, result):
        # Marshal data in result for passing to standard callback.
        # This routine is for results that contain status only, no data.
        if isinstance(result, Exception):
            status = result
        else:
            status = _AviaryCommon._get_status(result)
        return (status, None)

    @classmethod
    def _pretty_result(cls, result, host):
        if isinstance(result, urllib2.URLError):
            return Exception("Trouble reaching host %s, %s" % (host, result.reason))
        return result

    def _call_sync(self, process_results, meth, *meth_args, **meth_kwargs):
            # Common interface with QMF operations requires that a MethodResult
            # object (or something just like it) be returned for synchronous calls.
            # Might as well use the CallSync since the completion() routine 
            # handles status semantics.
            sync = CallSync()
            try:
                result = meth(*meth_args, **meth_kwargs)
            except Exception, e:
                result = e
            cb_args = process_results(result)
            sync.get_completion()(*cb_args) 
            return sync

class AviaryOperations(_AviaryCommon, _AviaryJobMethods, _AviaryQueryMethods):
    def __init__(self, name, datadir, job_servers, query_servers, 
                 key="", cert="", root_cert="", domain_verify=True):

        super(AviaryOperations, self).__init__(name, key, cert, root_cert, 
                                               domain_verify)

        _AviaryJobMethods.init(self, job_servers, datadir)
        _AviaryQueryMethods.init(self, query_servers, datadir)


class AviaryJobOperations(_AviaryCommon, _AviaryJobMethods):
    def __init__(self, name, datadir, job_servers, 
                 key="", cert="", root_cert="", domain_verify=True):

        super(AviaryJobOperations, self).__init__(name, key, cert, root_cert, 
                                                  domain_verify)
        
        _AviaryJobMethods.init(self, job_servers, datadir)

class AviaryQueryOperations(_AviaryCommon, _AviaryQueryMethods):
    def __init__(self, name, datadir, query_servers, 
                 key="", cert="", root_cert="", domain_verify=True):

        super(AviaryQueryOperations, self).__init__(name, key, cert, root_cert, 
                                                    domain_verify)

        _AviaryQueryMethods.init(self, query_servers, datadir)

def AviaryOperationsFactory(name, datadir, job_servers=None, query_servers=None,
                            key="", cert="", root_cert="", domain_verify=True):

    if job_servers and query_servers:
        return AviaryOperations(name, datadir, job_servers, query_servers,
                                key, cert, root_cert, domain_verify)
    elif job_servers:
        return AviaryJobOperations(name, datadir, job_servers,
                                   key, cert, root_cert, domain_verify)
    elif query_servers:
        return AviaryQueryOperations(name, datadir, query_servers,
                                   key, cert, root_cert, domain_verify)
    return None

try:
    # Some of this stuff does not exist pre suds 0.4.1
    # Make it work anyway for testing on such hosts by
    # declaring simple OverridesPlugin in the exception case.
    from suds.plugin import MessagePlugin
    from suds.sax.attribute import Attribute
    class OverridesPlugin(MessagePlugin):
        '''
        Plugin which allows optional addition of attributes
        to the suds message after marshalling.
        '''
        def __init__(self):
            self.set_attributes = False
            self.attributes = dict()

        def marshalled(self, context):
            if self.set_attributes:
                sj_body = context.envelope.getChild('Body')[0]
                for k,v in self.attributes.iteritems():
                    sj_body.attributes.append(Attribute(k, v))

except:
    class OverridesPlugin(object):
        def __init__(self):
            self.set_attributes = False

class OverrideClient(Client):
    '''
    Instantiate a Client object with an OverridesPlugin and
    retain a pointer to the plugin so that it can be controlled.
    '''
    def __init__(self, *args):
        self.override = OverridesPlugin()
        try:
            super(OverrideClient, self).__init__(plugins=[self.override], *args)
        except:
            super(OverrideClient, self).__init__(*args)

    def set_enable_attributes(self, truth):
        '''
        Set the flag that controls whether extra attributes are marshalled.

        The attribute set is specified through the set_attributes() method.
        If 'truth' is True, then service calls made through this client
        will have the extra attributes appended to the attribute set of service
        request.
        '''
        self.override.set_attributes = truth

    def set_attributes(self, attrs):
        '''
        Specify extra attributes to append to the attribute set
        of service requests made through this client.

        The extra attributes will be added if set_enable_attributes()
        has been called with a value of True sometime prior to the
        time of a request.
        '''
        self.override.attributes = attrs

class JobClientPool(ObjectPool):
    def __init__(self, job_wsdl, max_size):
        super(JobClientPool, self).__init__(max_size)
        self.job_wsdl = job_wsdl

    def create_object(self):
        return OverrideClient(self.job_wsdl)

class QueryClientPool(ObjectPool):
    def __init__(self, query_wsdl, max_size):
        super(QueryClientPool, self).__init__(max_size)
        self.query_wsdl = query_wsdl

    def create_object(self):
        return OverrideClient(self.query_wsdl)
