from parameters import ListParameter
from util import *
from widgets import ItemSet
from wooly import *

strings = StringCatalog(__file__)

class WidgetPage(Page):
    def __init__(self, app, name):
        super(WidgetPage, self).__init__(app, name)

        self.widget_classes = set()

        self.page_frame = self.FrameParameter(app, "frame")
        self.add_parameter(self.page_frame)

        self.set_default_frame(self)

        self.__saved_params = dict()

    def init_widget(self, widget):
        super(WidgetPage, self).init_widget(widget)

        for cls in widget.__class__.__mro__:
            self.widget_classes.add(cls)

            if cls is Widget:
                break

    def set_frame(self, session, frame):
        self.page_frame.set(session, frame)
        return frame

    def get_frame(self, session):
        return self.page_frame.get(session)

    def pop_frame(self, session):
        frame = self.get_frame(session)
        self.set_frame(session, frame.frame)
        return frame

    def set_default_frame(self, frame):
        self.page_frame.default = frame

    def get_page_parameters(self, session):
        """
        Gets a list of parameters saved for this page and its current
        frame.

        This, combined with [gs]et_current_frame, serves to discard
        state that is out of scope.  The current pattern is to
        preserve the state of parameters in the current frame and all
        of its ancestor frames.
        """

        frame = self.get_frame(session)

        try:
            params = self.__saved_params[frame]
        except KeyError:
            params = list()
            self.save_parameters(session, params)
            self.__saved_params[frame] = params

        return params

    class FrameParameter(Parameter):
        def do_marshal(self, frame):
            return frame.path

        def do_unmarshal(self, path):
            return self.widget.page.page_widgets_by_path.get(path)

class HtmlPage(WidgetPage):
    def __init__(self, app, name):
        super(HtmlPage, self).__init__(app, name)

        self.base_name = os.path.splitext(name)[0]

        self.updates = self.UpdatesAttribute(app, "updates")
        self.add_attribute(self.updates)

        self.defers = self.UpdatesAttribute(app, "defers")
        self.add_attribute(self.defers)

        self.update_script = UpdateScript(app, "update_script", self)
        self.add_child(self.update_script)

        self.defer_script = DeferScript(app, "defer_script", self)
        self.defer_script.interval = 0
        self.add_child(self.defer_script)

        messages = HtmlPageMessages(app, "messages")
        self.add_child(messages)

        self.update_page = UpdatePage(app, self.base_name + ".update", self)
        self.app.add_page(self.update_page)

        self.popup_page = PopupPage(app, self.base_name + ".popup", self)
        self.app.add_page(self.popup_page)

        self.css_page = CssPage(app, self.base_name + ".css", self)
        self.app.add_page(self.css_page)

        self.javascript_page = JavascriptPage \
            (app, self.base_name + ".js", self)
        self.app.add_page(self.javascript_page)

    def get_content_type(self, session):
        value = session.request_environment.get("HTTP_USER_AGENT")

        if (value and value.find("MSIE 6") != -1) or (value and value.find("MSIE 9") != -1):
            # sadly, although msie 9 handles application/xhtml+xml, the version of swfobject (2.2)
            # that we are using does not support that content type (by their design), so we fall back to text/html
            content_type = self.html_content_type
        else:
            content_type = self.xhtml_content_type

        return content_type

    def enable_update(self, session, widget):
        #print "Enabling update on widget %s" % widget

        self.updates.get(session).append(widget)

    def enable_defer(self, session, widget):
        self.defers.get(session).append(widget)

    def get_update_url(self, session, widgets):
        sess = Session(self.page.update_page)

        self.page.update_page.widgets.set(sess, widgets)
        self.page.update_page.session.set(sess, session)

        return sess.marshal()

    def get_popup_url(self, session, widget):
        sess = Session(self.page.popup_page)

        self.page.popup_page.widgets.set(sess, [widget])
        self.page.popup_page.session.set(sess, session)

        return sess.marshal()

    def do_process(self, session):
        self.update_script.process(session)
        self.defer_script.process(session)
        super(HtmlPage, self).do_process(session)

    def render_base_name(self, session):
        return self.base_name

    class UpdatesAttribute(Attribute):
        def get_default(self, session):
            return list()

class HtmlPageMessages(ItemSet):
    def do_get_items(self, session):
        return session.messages

class AjaxScript(Widget):
    def __init__(self, app, name, html_page):
        super(AjaxScript, self).__init__(app, name)

        self.html_page = html_page
        try:
            self.interval = app.update_interval * 1000
        except AttributeError:
            self.interval = 10000

    def do_render(self, session):
        if self.get_widget_list(session):
            return super(AjaxScript, self).do_render(session)

    def render_url(self, session):
        widgets = self.get_widget_list(session)
        return self.html_page.get_update_url(session, widgets)

    def get_widget_list(self, session):
        pass

    def render_interval(self, session):
        return self.interval

class DeferScript(AjaxScript):
    def get_widget_list(self, session):
        return self.html_page.defers.get(session)

class UpdateScript(AjaxScript):
    def get_widget_list(self, session):
        return self.html_page.updates.get(session)

class UpdatePage(Page):
    def __init__(self, app, name, html_page):
        super(UpdatePage, self).__init__(app, name)

        self.widget_tmpl = WidgetTemplate(self, "widget_html")

        self.html_page = html_page

        self.session = self.SessionParameter(app, "session")
        self.add_parameter(self.session)

        item = self.WidgetParameter(app, "item", self.html_page)

        self.widgets = ListParameter(app, "widget", item)
        self.add_parameter(self.widgets)

    def get_content_type(self, session):
        return self.xml_content_type

    def get_cache_control(self, session):
        return "no-cache"

    def do_process(self, session):
        # get the session for the main page
        sess = self.session.get(session)
        self.html_page.process(sess)
        # if the main page set it's redirect attribute in it's session
        redirect = self.html_page.redirect.get(sess)
        # pass that redirect on to the the update page
        if redirect:
            self.redirect.set(session, redirect)

    def render_widgets(self, session):
        writer = Writer()
        sess = self.session.get(session)

        widgets = self.widgets.get(session)

        for widget in widgets:
            self.widget_tmpl.render(writer, sess, widget)

        return writer.to_string()

    def render_widget_id(self, session, widget):
        return widget.path

    def render_widget(self, session, widget):
        return widget.render(session)

    class SessionParameter(Parameter):
        def do_marshal(self, session):
            return session.marshal()

        def do_unmarshal(self, string):
            return Session.unmarshal(self.app, string)

        def get(self, session):
            sess = super(UpdatePage.SessionParameter, self).get(session)

            sess.client_session = session.client_session
            sess.background = True

            return sess

    class WidgetParameter(Parameter):
        def __init__(self, app, name, page):
            self.page = page

        def do_marshal(self, widget):
            return widget.path

        def do_unmarshal(self, path):
            return self.page.page_widgets_by_path.get(path)

class PopupPage(UpdatePage):
    def get_content_type(self, session):
        return self.html_content_type

class CssPage(Page):
    def __init__(self, app, name, html_page):
        super(CssPage, self).__init__(app, name)

        self.html_page = html_page
        self.__then = datetime.utcnow()
        self.__css = None

    def get_last_modified(self, session):
        return self.__then

    def get_content_type(self, session):
        return "text/css"

    def get_cache_control(self, session):
        return "max-age=86400"

    def get_css(self):
        if not self.__css:
            writer = Writer()

            for cls in sorted(self.html_page.widget_classes):
                strs = cls.get_module_strings()

                if strs:
                    css = strs.get(cls.__name__ + ".css")

                    if css:
                        writer.write("/** %s.%s **/\r\n" % \
                                         (cls.__module__, cls.__name__))
                        writer.write("\r\n")
                        writer.write(css)
                        writer.write("\r\n")
                        writer.write("\r\n")

            self.__css = writer.to_string()

        return self.__css

    def do_render(self, session, *args):
        return self.get_css()

class JavascriptPage(Page):
    def __init__(self, app, name, html_page):
        super(JavascriptPage, self).__init__(app, name)

        self.html_page = html_page
        self.__then = datetime.utcnow()
        self.__javascript = None

    def get_last_modified(self, session):
        return self.__then

    def get_content_type(self, session):
        return "text/javascript"

    def get_cache_control(self, session):
        return "max-age=86400"

    def get_javascript(self):
        if not self.__javascript:
            writer = Writer()

            for cls in sorted(self.html_page.widget_classes):
                strs = cls.get_module_strings()

                if strs:
                    javascript = strs.get(cls.__name__ + ".javascript")

                    if javascript:
                        writer.write("/** %s.%s **/\r\n" % \
                                         (cls.__module__, cls.__name__))
                        writer.write("\r\n")
                        writer.write(javascript)
                        writer.write("\r\n")
                        writer.write("\r\n")

            self.__javascript = writer.to_string()

        return self.__javascript

    def do_render(self, session, *args):
        return self.get_javascript()

class ResourcePage(Page):
    content_types_by_extension = {
        ".css": "text/css",
        ".gif": "image/gif",
        ".html": "text/html",
        ".jpeg": "image/jpeg",
        ".jpg": "image/jpeg",
        ".js": "text/javascript",
        ".png": "image/png",
        }

    def __init__(self, app, name):
        super(ResourcePage, self).__init__(app, name)

        self.rname = Parameter(app, "name")
        self.add_parameter(self.rname)

        self.then = datetime.utcnow()

    def get_last_modified(self, session):
        return self.then

    def get_cache_control(self, session):
        return "max-age=86400"

    def get_content_type(self, session):
        name = self.rname.get(session)
        base, ext = os.path.splitext(name)
        ext = ext.lower()
        content_type = self.content_types_by_extension.get(ext, "text/plain")

        return content_type

    def do_render(self, session, *args):
        name = self.rname.get(session)

        if name:
            if name.startswith("/"):
                name = name[1:]

            resource = self.app.get_resource(name)

            if resource:
                return resource.read()

class CsvPage(WidgetPage):
    def get_content_type(self, session):
        return "text/csv"

    def get_file_name(self, session):
        return "cumin.csv"

    def get_extra_headers(self, session):
        file_name = self.get_file_name(session)
        return [('Content-Disposition', "attachment; filename=\"%s\"" % file_name), 
                ("Content-description", "File Transfer")]

    def do_render(self, session, *args):
        csv = self.render_content(session, *args)
        return csv and csv or " "
