from util import *
from wooly import *

class DictParameter(Parameter):
    def __init__(self, app, name):
        super(DictParameter, self).__init__(app, name)

        self.default = dict()
        self.is_dictionary = True

    @classmethod
    def sep(cls):
        return "|"

    def add(self, session, value, full_key):
        # full_key looks like <DictParameter.path>_<stuff>

        keys = self.get(session)

        # separate the DictParameter.path from the stuff
        foo, stuff = full_key.split(DictParameter.sep(), 1)
        if stuff:
            self.split_stuff(keys, stuff, value)

    def split_stuff(self, keys, stuff, value):
        """ Each time there is a delimiter in stuff,
            create a nested dictionary entry """

        # can't use: stuff_1, stuff_2 = stuff.split(DictParameter.sep(), 1)
        stuff_list = stuff.split(DictParameter.sep(), 1)
        stuff_1 = stuff_list[0]
        if len(stuff_list) > 1:
            stuff_2 = stuff_list[1]
            if not stuff_1 in keys:
                keys[stuff_1] = dict()
            self.split_stuff(keys[stuff_1], stuff_2, value)
        else:
            if value is not None:
                keys[stuff_1] = value
            elif stuff_1 in keys:
                del keys[stuff_1]

    def get_instance_key(self, key):
        return DictParameter.sep().join((self.path, key))

    def get_default(self, session):
        return copy(self.default)

    def clear(self):
        self.default = dict()

class ListParameter(Parameter):
    def __init__(self, app, name, item_parameter):
        super(ListParameter, self).__init__(app, name)

        self.item_parameter = item_parameter
        self.default = list()

        self.is_collection = True

    def get_default(self, session):
        return copy(self.default)

    def add(self, session, value, key=None):
        lst = self.get(session)
        lst.append(value)
        return lst

    def do_unmarshal(self, string):
        return self.item_parameter.do_unmarshal(string)

    def do_marshal(self, object):
        return self.item_parameter.do_marshal(object)

class StringParameter(Parameter):
    pass

# XXX this will be a strict sort of string parameter
class SymbolParameter(StringParameter):
    pass

class IntegerParameter(Parameter):
    def do_unmarshal(self, string):
        try:
            return int(string)
        except:
            return string

class FloatParameter(Parameter):
    def do_unmarshal(self, string):
        try:
            return float(string)
        except:
            return string

class BooleanParameter(Parameter):
    def __init__(self, app, name):
        Parameter.__init__(self, app, name)

        self.default = False

    def do_unmarshal(self, string):
        return string == "t"

    def do_marshal(self, object):
        return object and "t" or "f"

class VoidBooleanParameter(Parameter):
    def get(self, session):
        return self.path in session.values_by_path

    def set(self, session, value):
        key = self.path

        if value:
            session.set(key, None)
        else:
            session.unset(key)

class PageParameter(Parameter):
    def __init__(self, app, name):
        super(PageParameter, self).__init__(app, name)

    def do_unmarshal(self, string):
        return self.app.pages_by_name.get(string)

    def do_marshal(self, page):
        return page.name
